 /*
 * Copyright (c) 2023-2024 Felix Ingrand, CNRS/LAAS
 *
 * All rights reserved.
 * 
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source  code must retain the  above copyright
 *    notice and this list of conditions.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice and  this list of  conditions in the  documentation and/or
 *    other materials provided with the distribution.
 * 
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "drone3.h"
#include <unistd.h>
#include <pthread.h>
#include <limits.h>

#include "bt2f-hippo.h"
#include "drone3-action-arg.h"

#include "h2errorLib.h"
#include "pom/c/client.h"
#include "maneuver/c/client.h"
#include "rotorcraft/c/client.h"

//typedef int bool;

#define false 0
#define true 1

#include "hippo_app_functions.h"

#define DRAW(max) (!(rand() % (max))) /* Roll a max dice, 1 chance out of max. */

/* Sleep a random amount of time between the 2 values in seconds. */
#define RANDOM_SLEEP(min, max)						\
  do {									\
    struct timespec to_sleep, remain_time;				\
    double diff = min + (max - min)*(double)rand() / (double)RAND_MAX;	\
    to_sleep.tv_sec = floor(diff);					\
    to_sleep.tv_nsec = (diff - floor(diff)) * 1000000000;		\
    while (nanosleep(&to_sleep, &remain_time)) {			\
      to_sleep.tv_sec = remain_time.tv_sec;				\
      to_sleep.tv_nsec = remain_time.tv_nsec;				\
    }									\
  } while (0)


extern int get_fls_value(void);

// pthread_cond_t producer, consumer; /* event producer (external function) and event consumer (the fiacre model) */
// pthread_mutex_t mp;
// pthread_t pom_checker, battery_checker;

/* #define MAX_EVT_QUEUE 10 */
/* drone_event_type_c event_queued[MAX_EVT_QUEUE]; */

/* int front_index = 0;		/\* the index of the first free slot *\/ */
/* int back_index = 0;		/\* the index the last free slot *\/ */
/* int occupied = 0;  */



//void queue_event(drone_event_type_c event);

//typedef enum {Low, Critical, Good} sv_battery_c;
//typedef enum {Invalid, Coarse, Valid} sv_localization_status_c;


/* void *check_localization_status (void *ignore); */
/* void *check_battery (void *ignore); */

/* void init_event_handler(void) */
/* { */
/*   pthread_mutex_init(&mp, NULL); */
/*   pthread_cond_init(&producer, NULL); */
/*   pthread_cond_init(&consumer, NULL); */

/*   int rc; */
/*   rc = pthread_create(&pom_checker, NULL, check_localization_status, NULL); */
/*   if (rc ){ */
/*     fprintf(stderr,"init_event_handler: ERROR; pthread_create() check_localization_status, returned %d\n", rc); */
/*     perror(NULL); */
/*   } */

/*   rc = pthread_create(&battery_checker, NULL, check_battery, NULL); */
/*   if (rc ){ */
/*     fprintf(stderr,"init_event_handler: ERROR; pthread_create() check_battery, returned %d\n", rc); */
/*     perror(NULL); */
/*   } */
/* } */

static genom_client rotorcraft_genom_client = NULL;

void create_rotorcraft_client_if_needed(void)
{
  if (!rotorcraft_genom_client) {
    genom_event ex;
    const void *exdetail;
    int argc = 0;
    char *argv[1];
    
    argv[0] = NULL;		/*  is this really needed? */
    rotorcraft_genom_client = genom_rotorcraft_client_init(argc,argv,&ex,&exdetail);
    
    if (!rotorcraft_genom_client) {
      uc_log_info("genom_rotorcraft_client_init: failed, %s, %s", ex, (char*)exdetail);
      h2perror(NULL);
    }
  }
}

static int get_battery_rqstid = -1;

bool Fiacre_Action_start_drone_halt_c(btnode_record_c arg)
{
  return true;
}

ret_status_c Fiacre_Action_start_drone_c(btnode_record_c arg)
{

  if DRAW(3) return running;
  
  if DRAW(15) return failure;
  
  return success;

}
static int stop_rqstid = -1;

bool Fiacre_Action_shutdown_drone_halt_c(btnode_record_c arg)
{
  if (stop_rqstid == -1) {
    uc_log_info("Fiacre_landing_interrupt_action_c: stop service not running.");
    return false;
  } else {
    genom_event res, report;

    if  (!rotorcraft_genom_client) {
      uc_log_info("Fiacre_landing_interrupt_action_c: no rotorcraft_genom_client.");
      return false;
    }
  
    struct genom_rotorcraft_abort_activity_input input;
    genom_rotorcraft_client_abort_activity_init_input(&input);
  
    struct genom_rotorcraft_abort_activity_output *output;
  
    input.activity = stop_rqstid;

    int rqstid;
 
    res = genom_rotorcraft_client_abort_activity_rqst(rotorcraft_genom_client, &input, NULL, NULL, NULL, &rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_rotorcraft_client_abort_activity_rqst: %s", res);
    }

    res = genom_rotorcraft_client_wait(rotorcraft_genom_client, rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_rotorcraft_client_wait: %s", res);
    }

    void *exdetail2;
    res = genom_rotorcraft_client_abort_activity_result(rotorcraft_genom_client, rqstid, &report, &output, &exdetail2);
  
    genom_rotorcraft_client_abort_activity_fini_input(&input);

    if (res != genom_ok) {
      uc_log_info("genom_rotorcraft_client_abort_activity_result: res: %s", res);
    }

    if (report == genom_ok) {
      return true;
    } else {
      uc_log_info("genom_rotorcraft_client_abort_activity_result: report %s", report);
      return false;
    }
  }
  return true;			/* to be improved... (abort the stop...) */
}

ret_status_c Fiacre_Action_shutdown_drone_c(btnode_record_c arg)
{
  genom_event res, report;

  create_rotorcraft_client_if_needed();
  struct genom_rotorcraft_stop_input input;

  if  (!rotorcraft_genom_client) return failure;

  if (stop_rqstid == -1) {	/* nothing running, launch it */

    genom_rotorcraft_client_stop_init_input(&input);

    //shutdown_drone_action_arg(arg.ArgIndex);
     
    res = genom_rotorcraft_client_stop_rqst(rotorcraft_genom_client, &input, NULL, NULL, NULL, &stop_rqstid);
    
    if (res != genom_ok) {
      uc_log_info("genom_rotorcraft_client_stop_rqst: %s", res);
      return failure;
    }
    
    return running;
    
  } else { 
    
    struct genom_rotorcraft_stop_output *output;
    int done;

    done = genom_rotorcraft_client_done(rotorcraft_genom_client, stop_rqstid);

    uc_log_info("genom_rotorcraft_client_stop done: %d", done);
    
    if (!done) return running;

    void *exdetail2;
    res = genom_rotorcraft_client_stop_result(rotorcraft_genom_client, stop_rqstid, &report, &output, &exdetail2);
    
    stop_rqstid = -1;
    
    genom_rotorcraft_client_stop_fini_input(&input);
    
    if (res != genom_ok) {
      uc_log_info("genom_rotorcraft_client_stop_result: res: %s", res);
    }

    if (report == genom_ok) {
      return success;
    } else {
      uc_log_info("genom_rotorcraft_client_stop_result: report %s", report);
      //queue_event(flight_status_to_in_air_unsafe_event);
      return failure;
    }
  }
}

static genom_client maneuver_genom_client = NULL;

void create_maneuver_client_if_needed(void)
{
  if (!maneuver_genom_client) {
    genom_event ex;
    const void *exdetail;
    int argc = 0;
    char *argv[1];
    
    argv[0] = NULL;		/*  is this really needed? */
    maneuver_genom_client = genom_maneuver_client_init(argc,argv,&ex,&exdetail);
    
    if (!maneuver_genom_client) {
      uc_log_info("genom_maneuver_client_init: failed, %s, %s", ex, (char*)exdetail);
      h2perror(NULL);
    }
  }
}


static int take_off_rqstid = -1;

ret_status_c Fiacre_Action_takeoff_c(btnode_record_c arg)
{
  genom_event res, report;
  struct genom_maneuver_take_off_input input;

  create_maneuver_client_if_needed();

  if  (!maneuver_genom_client) return failure; //takeoff_failure_grounded;


  if (take_off_rqstid == -1) {	/* nothing running, launch it */

    genom_maneuver_client_take_off_init_input(&input);
  
    takeoff_action_arg(arg.ArgIndex);

    res = genom_maneuver_client_take_off_rqst(maneuver_genom_client, &input, NULL, NULL, NULL, &take_off_rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_take_off_rqst: %s", res);
      return failure;
    }

    return running;
    
  } else { 
    
    struct genom_maneuver_take_off_output *output;
    int done;
    
    done = genom_maneuver_client_done(maneuver_genom_client, take_off_rqstid);
    
    uc_log_info("genom_maneuver_client_take_off done: %d", done);

    if (! done) return running;
    
    void *exdetail2;
    res = genom_maneuver_client_take_off_result(maneuver_genom_client, take_off_rqstid, &report, &output, &exdetail2);

    //    uc_log_info("genom_maneuver_client_take_off_result: res %s, report %s", res, report);
    
    take_off_rqstid = -1;
    
    genom_maneuver_client_take_off_fini_input(&input);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_take_off_result: res: %s", res);
    }

    if (report == genom_ok) {
      return success; // takeoff_success_at_altitude;
    } else if (report == genom_interrupted_id) {
      return failure; // takeoff_interrupted;
    } else {
      uc_log_info("genom_maneuver_client_take_off_result: report %s", report);
      // queue_event(flight_status_to_in_air_unsafe_event);
      return failure; // takeoff_failure_grounded;
    }
  }
}

static int velocity_rqstid = -1;

ret_status_c Fiacre_set_velocity_action_c(btnode_record_c arg)
{
  genom_event res, report;
  struct genom_maneuver_velocity_input input;

  create_maneuver_client_if_needed();

  if  (!maneuver_genom_client) return failure; // set_velocity_failure_landing_pb;

  if (velocity_rqstid == -1) {	/* nothing running, launch it */
    
    genom_maneuver_client_velocity_init_input(&input);
    
    input.vx = 0.;
    input.vy = 0.;
    input.vz = 0.;
    input.wz = 0.;
    
    input.ax = 0.;
    input.ay = 0.;
    input.az = 0.;
    input.duration = 0.;
    
    //    set_velocity_action_arg(arg.ArgIndex);

    res = genom_maneuver_client_velocity_rqst(maneuver_genom_client, &input, NULL, NULL, NULL, &velocity_rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_velocity_rqst: %s", res);
      return failure;
    }

    return running;
    
  } else {

    struct genom_maneuver_velocity_output *output;
    int done;
    
    done = genom_maneuver_client_done(maneuver_genom_client, velocity_rqstid);

    uc_log_info("genom_maneuver_client_set_velocity done: %d", done);

    if (! done) return running;

    void *exdetail2;
    res = genom_maneuver_client_velocity_result(maneuver_genom_client, velocity_rqstid, &report, &output, &exdetail2);

    velocity_rqstid = -1;
  
    genom_maneuver_client_velocity_fini_input(&input);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_velocity_result: res: %s", res);
    }

    if (report == genom_ok) {
      return success; // set_velocity_success_drone_descending;
    } else {
      uc_log_info("genom_maneuver_client_velocity_result: report %s", report);
      return failure; // set_velocity_failure_landing_pb;
    }
  }
}

bool Fiacre_Action_land_halt_c(btnode_record_c arg)
{
  if (take_off_rqstid == -1) {
    uc_log_info("Fiacre_Action_land_halt_c: take_off service not running.");
    return false;
  } else {
    genom_event res, report;

    if  (!maneuver_genom_client) {
      uc_log_info("Fiacre_Action_land_halt_c: no maneuver_genom_client.");
      return false;
    }
  
    struct genom_maneuver_abort_activity_input input;
    genom_maneuver_client_abort_activity_init_input(&input);
  
    struct genom_maneuver_abort_activity_output *output;
  
    input.activity = take_off_rqstid;

    int rqstid;
 
    res = genom_maneuver_client_abort_activity_rqst(maneuver_genom_client, &input, NULL, NULL, NULL, &rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_abort_activity_rqst: %s", res);
    }

    res = genom_maneuver_client_wait(maneuver_genom_client, rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_wait: %s", res);
    }

    void *exdetail2;
    res = genom_maneuver_client_abort_activity_result(maneuver_genom_client, rqstid, &report, &output, &exdetail2);
  
    genom_maneuver_client_abort_activity_fini_input(&input);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_abort_activity_result: res: %s", res);
    }

    if (report == genom_ok) {
      return true;
    } else {
      uc_log_info("genom_maneuver_client_abort_activity_result: report %s", report);
      return false;
    }
  }
  return true;			/* need to be properly done */
}

ret_status_c Fiacre_Action_land_c(btnode_record_c arg)
{
  genom_event res, report;
  struct genom_maneuver_take_off_input input;

  create_maneuver_client_if_needed();

  if  (!maneuver_genom_client) return failure; // landing_failure_aborted;

  if (take_off_rqstid == -1) {	/* nothing running, launch it */
  
    genom_maneuver_client_take_off_init_input(&input);
  
    input.height = 0.0;
    input.duration = 0.0;
    
    res = genom_maneuver_client_take_off_rqst(maneuver_genom_client, &input, NULL, NULL, NULL, &take_off_rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_take_off_rqst: %s", res);
      return failure;
    }

    return running;
  } else {
    struct genom_maneuver_take_off_output *output;
    int done;
  
    done = genom_maneuver_client_done(maneuver_genom_client, take_off_rqstid);

    uc_log_info("genom_maneuver_client_take_off done: %d", done);

    if (! done) return running;
    
    void *exdetail2;
    res = genom_maneuver_client_take_off_result(maneuver_genom_client, take_off_rqstid, &report, &output, &exdetail2);
    
    take_off_rqstid = -1;
    genom_maneuver_client_take_off_fini_input(&input);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_take_off_result: res %s", res);
    }

    if (report == genom_ok) { 
      //    queue_event(flight_status_to_on_ground_event);
      return success; // landing_success_on_ground;
    } else if (report == genom_interrupted_id) {
      return failure; // landing_interrupted;
    } else {
      uc_log_info("genom_maneuver_client_take_off_result: report %s", report);
      // queue_event(flight_status_to_in_air_unsafe_event);
      return failure; // landing_failure_emergency;
    }
  }
}
static int goto_rqstid = -1;

ret_status_c Fiacre_Action_goto_waypoint_c(btnode_record_c arg)
{
  genom_event res, report;

  create_maneuver_client_if_needed();

  if  (!maneuver_genom_client) return failure; // goto_waypoint_failure_not_arrived;

  struct genom_maneuver_goto_input input;

  if (goto_rqstid == -1) {	/* nothing running, launch it */
  
    genom_maneuver_client_goto_init_input(&input);
  
    goto_waypoint_action_arg(arg.ArgIndex);
    
    res = genom_maneuver_client_goto_rqst(maneuver_genom_client, &input, NULL, NULL, NULL, &goto_rqstid);
    
    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_goto_rqst: %s", res);
      return failure;
    }

    return running;
    
  } else {
    struct genom_maneuver_goto_output *output;
    int done;
    
    done = genom_maneuver_client_done(maneuver_genom_client, goto_rqstid);

    uc_log_info("genom_maneuver_client_goto done: %d", done);

    if (! done) return running;

    void *exdetail2;
    res = genom_maneuver_client_goto_result(maneuver_genom_client, goto_rqstid, &report, &output, &exdetail2);

    goto_rqstid = -1;
    
    genom_maneuver_client_goto_fini_input(&input);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_goto_result: res %s", res);
    }

    if (report == genom_ok) {
      return success; // goto_waypoint_success_arrived;
    } else if (report == genom_interrupted_id) {
      return failure; // goto_waypoint_interrupted;
    } else {
      uc_log_info("genom_maneuver_client_goto_result: report %s", report);
      
      return failure; //goto_waypoint_failure_not_arrived;
    }
  }
}

/* drone_event_type_c Fiacre_drone_event_c(void) */
/* { */
/*   h_log_info("Fiacre_drone_event_c: Arming event port"); */

/*   //  RANDOM_SLEEP(0.05,0.05);	/\* this is to address the pb that multiple events on the same port  */
/*   //				   in one engine loop are not properly handled. *\/ */

/*   pthread_mutex_lock(&mp); */
  
/*   while (occupied == 0) { */
/*     h_log_info("Fiacre_drone_event_c: Wait for events"); */
/*     pthread_cond_wait(&consumer, &mp); */
/*   } */
/*   h_log_info("Fiacre_drone_event_c: Wait for events is over"); */

/*   int index = back_index;			/\* in case the var gets modified outside of the lock *\/ */

/*   back_index = (back_index + 1) % MAX_EVT_QUEUE; */
/*   occupied--; */
  
/*   h_log_info("Fiacre_drone_event_c: Signaling the producer"); */
/*   pthread_cond_signal(&producer); */

/*   pthread_mutex_unlock(&mp); */
/*   h_log_info("Fiacre_drone_event_c: Returning %d in position %d in queue from event port", event_queued[index], index); */
  
/*   return event_queued[index]; */
/* } */

/* void queue_event(drone_event_type_c event) */
/* { */
/*   h_log_info("queue_event: Queueing event %d", event); */
/*   pthread_mutex_lock(&mp); */

/*   while (occupied == MAX_EVT_QUEUE) { */
/*     h_log_info("queue_event: Wait for space in the queue."); */
/*     pthread_cond_wait(&producer, &mp); */
/*   } */
/*   h_log_info("queue_event: Wait for space is over"); */
  
/*   event_queued[front_index] = event; */
  
/*   h_log_info("queue_event: queueing %d in %d", event, front_index); */

/*   front_index = (front_index + 1) %  MAX_EVT_QUEUE; */
/*   occupied++; */
  
/*   h_log_info("queue_event: Signaling the consumer"); */
/*   pthread_cond_signal(&consumer); */
/*   pthread_mutex_unlock(&mp); */
/* } */

/* drone_interrupt_type_c Fiacre_drone_interrupt_c() */
/* { */
/*   sleep(6000); */

/*   return goto_waypoint_interrupt; */
/* } */



static genom_client pom_genom_client = NULL;

void create_pom_client_if_needed(void)
{
  if (!pom_genom_client) {
    genom_event ex;
    const void *exdetail;
    int argc = 0;
    char *argv[1];
    
    argv[0] = NULL;		/*  is this really needed? */
    pom_genom_client = genom_pom_client_init(argc,argv,&ex,&exdetail);
    
    if (!pom_genom_client) {
      uc_log_info("genom_pom_client_init: failed, %s, %s", ex, (char*)exdetail);
      h2perror(NULL);
    } 
  }
}

/* void *check_localization_status (void *ignore) */
/* { */
/*   do {     */
/*     create_pom_client_if_needed(); */
/*     sleep (1); */
/*   } while (!pom_genom_client); */

/*   sv_localization_status_c prev_status = Invalid; // Coarse Valid */
/*   sv_localization_status_c status; */
/*   //  queue_event(localization_status_to_valid_event); */
  
/*   //  RANDOM_SLEEP(0.5,0.5);		/\* let the engine start *\/ */

/*   // while (1) { */
/*     or_pose_estimator_state data; */
    
/*     genom_event res = genom_pom_client_frame_port(pom_genom_client, "robot", &data); */
    
/*     if (res != genom_ok) { */
/*       uc_log_info("check_localization_status: genom_pom_client_frame_port: %s", res); */
/*       status = Invalid; */
/*     } else { */

/*       if (data.pos_cov._present) { */
/* 	double hcov = sqrt((data.pos_cov._value.cov[0] /\* x diag *\/ */
/* 		       + data.pos_cov._value.cov[2]) /\* y diag *\/ */
/* 		       /2); */
/* 	double vcov = sqrt(data.pos_cov._value.cov[5]); /\* z diag *\/ */
	
/* 	//	uc_log_info("hvar: %g, vvar  %g", hcov, vcov); */
	
/* 	if (hcov < 0.01) */
/* 	  status = Valid; */
/* 	else if (hcov < 0.1) */
/* 	  status = Coarse; */
/* 	else */
/* 	  status = Invalid; */
/*       } else { */
/* 	status = Invalid; */
/*       } */
/*     } */
    
/*     if (status != prev_status) { */
/*       prev_status = status; */
/*       switch (status) { */
/*       case Invalid: */
/* 	uc_log_info("Localisation Invalid"); */
/* 	//queue_event(localization_status_to_invalid_event); */
/* 	break; */
/*       case Valid: */
/* 	uc_log_info("Localisation Valid"); */
/* 	//queue_event(localization_status_to_valid_event); */
/* 	break; */
/*       case Coarse: */
/* 	uc_log_info("Localisation Coarse"); */
/* 	//queue_event(localization_status_to_coarse_event); */
/* 	break; */
/*       } */
/*     } */
/*     //    RANDOM_SLEEP(0.5, 0.5); */
/*     //  } */

/*   return NULL; */
/* } */


bool Fiacre_Action_takeoff_halt_c(btnode_record_c arg)
{
  if (take_off_rqstid == -1) {
    uc_log_info("Fiacre_takeoff_interrupt_action_c: take_off service not running.");
    return false;
  } else {
    genom_event res, report;

    if  (!maneuver_genom_client) {
      uc_log_info("Fiacre_takeoff_interrupt_action_c: no maneuver_genom_client.");
      return false;
    }
  
    struct genom_maneuver_abort_activity_input input;
    genom_maneuver_client_abort_activity_init_input(&input);
  
    struct genom_maneuver_abort_activity_output *output;
  
    input.activity = take_off_rqstid;

    int rqstid;
 
    res = genom_maneuver_client_abort_activity_rqst(maneuver_genom_client, &input, NULL, NULL, NULL, &rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_abort_activity_rqst: %s", res);
    }

    res = genom_maneuver_client_wait(maneuver_genom_client, rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_wait: %s", res);
    }

    void *exdetail2;
    res = genom_maneuver_client_abort_activity_result(maneuver_genom_client, rqstid, &report, &output, &exdetail2);
  
    genom_maneuver_client_abort_activity_fini_input(&input);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_abort_activity_result: res: %s", res);
    }

    if (report == genom_ok) {
      return true;
    } else {
      uc_log_info("genom_maneuver_client_abort_activity_result: report %s", report);
      return false;
    }
  }
}

bool Fiacre_landing_interrupt_action_c()
{
  if (take_off_rqstid == -1) {
    uc_log_info("Fiacre_landing_interrupt_action_c: take_off service not running.");
    return false;
  } else {
    genom_event res, report;

    if  (!maneuver_genom_client) {
      uc_log_info("Fiacre_landing_interrupt_action_c: no maneuver_genom_client.");
      return false;
    }
  
    struct genom_maneuver_abort_activity_input input;
    genom_maneuver_client_abort_activity_init_input(&input);
  
    struct genom_maneuver_abort_activity_output *output;
  
    input.activity = take_off_rqstid;

    int rqstid;
 
    res = genom_maneuver_client_abort_activity_rqst(maneuver_genom_client, &input, NULL, NULL, NULL, &rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_abort_activity_rqst: %s", res);
    }

    res = genom_maneuver_client_wait(maneuver_genom_client, rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_wait: %s", res);
    }

    void *exdetail2;
    res = genom_maneuver_client_abort_activity_result(maneuver_genom_client, rqstid, &report, &output, &exdetail2);
  
    genom_maneuver_client_abort_activity_fini_input(&input);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_abort_activity_result: res: %s", res);
    }

    if (report == genom_ok) {
      return true;
    } else {
      uc_log_info("genom_maneuver_client_abort_activity_result: report %s", report);
      return false;
    }
  }
}

bool Fiacre_Action_goto_waypoint_halt_c(btnode_record_c arg) 
{
  if (goto_rqstid == -1) {
    uc_log_info("Fiacre_Action_goto_waypoint_halt_c: goto service not running.");
    return false;
  } else {
    genom_event res, report;

    if  (!maneuver_genom_client) {
      uc_log_info("Fiacre_Action_goto_waypoint_halt_c: no maneuver_genom_client.");
      return false;
    }
  
    struct genom_maneuver_abort_activity_input input;
    genom_maneuver_client_abort_activity_init_input(&input);
  
    struct genom_maneuver_abort_activity_output *output;
  
    input.activity = goto_rqstid;

    int rqstid;
 
    res = genom_maneuver_client_abort_activity_rqst(maneuver_genom_client, &input, NULL, NULL, NULL, &rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_abort_activity_rqst: %s", res);
    }

    res = genom_maneuver_client_wait(maneuver_genom_client, rqstid);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_wait: %s", res);
    }

    void *exdetail2;
    res = genom_maneuver_client_abort_activity_result(maneuver_genom_client, rqstid, &report, &output, &exdetail2);
  
    genom_maneuver_client_abort_activity_fini_input(&input);

    if (res != genom_ok) {
      uc_log_info("genom_maneuver_client_abort_activity_result: res: %s", res);
    }

    if (report == genom_ok) {
      return true;
    } else {
      uc_log_info("genom_maneuver_client_abort_activity_result: report %s", report);
      return false;
    }
  }
}

bool Fiacre_Action_start_camera_halt_c(btnode_record_c arg)
{
  return true;
}

ret_status_c Fiacre_Action_start_camera_c(btnode_record_c arg)
{
  if DRAW(3) return running;
  
  //  if DRAW(5) return failure;
  
  return success; // start_camera_success_camera_on;
}

static bool camera_tracking = false;

ret_status_c Fiacre_Action_camera_tracking_c(btnode_record_c arg)
{

  if (! camera_tracking) {
    camera_tracking = true;
  }
  
  if DRAW(120) {
      camera_tracking = false;
      return success;
    }
  
  if DRAW(240) {
      camera_tracking = false;
      return failure;
    }
  return running;
}

bool Fiacre_Action_camera_tracking_halt_c(btnode_record_c arg)
{
  camera_tracking = false;
}


sv_battery_c Fiacre_SetSV_measure_battery_c()
{
  sv_battery_c status = Good;

  return status;
}

ret_status_c Fiacre_Condition_battery_ok_c(btnode_record_c arg)
{
  bool simulation = false;
  float sim_level = 16.0;

  if DRAW(100) return failure;
  
  create_rotorcraft_client_if_needed();
  
  if  (!rotorcraft_genom_client) return failure;

  //  static sv_battery_c prev_status = Low; // Coarse Valid
  sv_battery_c status;
  
  genom_event res, report;
  double level;

  struct genom_rotorcraft_get_battery_input input;
  genom_rotorcraft_client_get_battery_init_input(&input);
  
  struct genom_rotorcraft_get_battery_output *output;
  
  if (get_battery_rqstid != -1) {
    uc_log_info("genom_rotorcraft_client_get_battery_rqst: already running, no reentrance.");
    return failure;
  }
  
  res = genom_rotorcraft_client_get_battery_rqst(rotorcraft_genom_client, &input, NULL, NULL, NULL, &get_battery_rqstid);

  if (res != genom_ok) {
    uc_log_info("genom_rotorcraft_client_get_battery_rqst: %s", res);
  }
  
  res = genom_rotorcraft_client_wait(rotorcraft_genom_client, get_battery_rqstid);

  if (res != genom_ok) {
    uc_log_info("genom_rotorcraft_client_wait: %s", res);
  }

  void *exdetail2;
  res = genom_rotorcraft_client_get_battery_result(rotorcraft_genom_client, get_battery_rqstid, &report, &output, &exdetail2);

  get_battery_rqstid = -1;
  
  genom_rotorcraft_client_get_battery_fini_input(&input);

  if (res != genom_ok) {
    uc_log_info("genom_rotorcraft_client_get_battery_result: res: %s", res);
  }

  if (report == genom_ok) {
    level = output->battery.level;
    // uc_log_info("Battery Level %f",level);
  } else {
    uc_log_info("genom_rotorcraft_client_get_battery_result: report %s", report);
    level = 0.0;
  }

  if (level >= 16.0) {	/* Hopefuly, we never reach this on real battery... */
    simulation = true;
    sim_level = sim_level - 0.05; 
    level = sim_level;
  }

  //    uc_log_info("Battery level %2f", level);
    
  if (level > 15.5)
    status = Good;
  else if (level > 14.8)
    status = Low;
  else
    status = Critical;
    
  /* if (status != prev_status) { */
  /*   prev_status = status; */
  switch (status) {
  case Critical:
    uc_log_info("Battery Critical%s", (simulation?" (sim)":""));
    return failure;
    //queue_event(battery_to_critical_event);
    break;
  case Good:
    uc_log_info("Battery Good%s", (simulation?" (sim)":""));
    //queue_event(battery_to_good_event);
    break;
  case Low:
    uc_log_info("Battery Low%s", (simulation?" (sim)":""));
    //queue_event(battery_to_low_event);
    break;
  }
  /* } */
  return success;
}

ret_status_c Fiacre_Condition_localization_ok_c(btnode_record_c arg)
{

  //  if DRAW(50) return failure;
   if DRAW(1000) return failure;

  create_pom_client_if_needed();

  if (!pom_genom_client) return failure;
  
  sv_localization_status_c status;
  or_pose_estimator_state data;
    
  genom_event res = genom_pom_client_frame_port(pom_genom_client, "robot", &data);
  
  if (res != genom_ok) {
    uc_log_info("check_localization_status: genom_pom_client_frame_port: %s", res);
    status = Invalid;
  } else {
    
    if (data.pos_cov._present) {
      double hcov = sqrt((data.pos_cov._value.cov[0] /* x diag */
			  + data.pos_cov._value.cov[2]) /* y diag */
			 /2);
      double vcov = sqrt(data.pos_cov._value.cov[5]); /* z diag */
      
      //	uc_log_info("hvar: %g, vvar  %g", hcov, vcov);
      
      if (hcov < 0.01)
	status = Valid;
      else if (hcov < 0.1)
	status = Coarse;
      else
	status = Invalid;
    } else {
      status = Invalid;
    }
  }
  
  switch (status) {
  case Invalid:
    uc_log_info("Localisation Invalid");
    //queue_event(localization_status_to_invalid_event);
    return failure;
    break;
  case Valid:
    uc_log_info("Localisation Valid");
    //queue_event(localization_status_to_valid_event);
    return success;
    break;
  case Coarse:
    uc_log_info("Localisation Coarse");
    //queue_event(localization_status_to_coarse_event);
    return failure;
    break;
  }
}





