 /*
 * Copyright (c) 2024-2024 Felix Ingrand, CNRS/LAAS
 *
 * All rights reserved.
 * 
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source  code must retain the  above copyright
 *    notice and this list of conditions.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice and  this list of  conditions in the  documentation and/or
 *    other materials provided with the distribution.
 * 
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <pthread.h>
#include <limits.h>

#include "bt2f-hippo.h"
#include "ros2-nav3-action-arg.h"

#include "ros2-nav3.h"

typedef int bool;

#define false 0
#define true 1

#include "hippo_app_functions.h"

#define DRAW(max) (!(rand() % (max))) /* Roll a max dice, 1 chance out of max. */

/* Sleep a random amount of time between the 2 values in seconds. */
#define RANDOM_SLEEP(min, max)						\
  do {									\
    struct timespec to_sleep, remain_time;				\
    double diff = min + (max - min)*(double)rand() / (double)RAND_MAX;	\
    to_sleep.tv_sec = floor(diff);					\
    to_sleep.tv_nsec = (diff - floor(diff)) * 1000000000;		\
    while (nanosleep(&to_sleep, &remain_time)) {			\
      to_sleep.tv_sec = remain_time.tv_sec;				\
      to_sleep.tv_nsec = remain_time.tv_nsec;				\
    }									\
  } while (0)


ret_status_c Fiacre_Condition_DistanceController_c(btnode_record_c arg)
{
  static int first = true;

  if (first) {
    first = false;
    return success;
  }
    
  if DRAW(50) return success;
  return failure;
}

ret_status_c Fiacre_Action_ComputePathToPose_c(btnode_record_c arg)
{
  if DRAW(5) return failure;
  return success;
}

ret_status_c Fiacre_Action_FollowPath_c(btnode_record_c arg)
{
  if DRAW(30) return failure;
  if DRAW(500) return success;
  return running;
}


ret_status_c Fiacre_Condition_GoalUpdated_c(btnode_record_c arg)
{
  static int first = true;

  if (first) {
    first = false;
    return success;
  }
    
  if DRAW(50) return success;
  return failure;
}

ret_status_c Fiacre_Action_ClearEntireCostmap_c(btnode_record_c arg)
{
  if DRAW(100) return failure;
  return success;
}

ret_status_c Fiacre_Action_Spin_c(btnode_record_c arg)
{
  if DRAW(10) return failure;
  return success;
}

ret_status_c Fiacre_Action_Wait_c(btnode_record_c arg)
{
  if DRAW(10) return failure;
  return success;
}

ret_status_c Fiacre_Action_BackUp_c(btnode_record_c arg)
{
  if DRAW(10) return failure;
  return success;
}

ret_status_c Fiacre_Action_ControllerSelector_c(btnode_record_c arg)
{
  if DRAW(100) return failure;
  return success;
}

ret_status_c Fiacre_Action_PlannerSelector_c(btnode_record_c arg)
{
  if DRAW(100) return failure;
  return success;
}

ret_status_c Fiacre_Condition_WouldAPlannerRecoveryHelp_c(btnode_record_c arg)
{
  if DRAW(30) return failure;
  return success;
}

ret_status_c Fiacre_Condition_WouldAControllerRecoveryHelp_c(btnode_record_c arg)
{
  if DRAW(30) return failure;
  return success;
}

bool  Fiacre_Action_ControllerSelector_halt_c(btnode_record_c arg)
{
  return success;
}

bool  Fiacre_Action_PlannerSelector_halt_c(btnode_record_c arg)
{
  return success;
}


bool Fiacre_Action_ClearEntireCostmap_halt_c(btnode_record_c arg)
{
  return true;
}

bool Fiacre_Action_Spin_halt_c(btnode_record_c arg)
{
  return true;
}

bool Fiacre_Action_Wait_halt_c(btnode_record_c arg)
{
  return true;
}

bool Fiacre_Action_BackUp_halt_c(btnode_record_c arg)
{
  return true;
}

bool Fiacre_Action_ComputePathToPose_halt_c(btnode_record_c arg)
{
  return true;
}

bool Fiacre_Action_FollowPath_halt_c(btnode_record_c arg)
{
  return true;
}


