/*                               -*- Mode: C -*- 
 * op-structure.h -- definition of op-structure
 * 
 * Copyright (c) 2024-2024 Francois Felix Ingrand.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef INCLUDE_bt_structure
#define INCLUDE_bt_structure

#include "shashPack_f.h"

#include "opaque.h"
#include "bt2f-type.h"
#include "bt2f-sprint.h"

#define BTTYPES T(UnknownBTNodeType)		\
       T(BehaviorTree)				\
       T(Action)				\
       T(Condition)				\
       T(Eval)					\
       T(SetSV)					\
       T(Sequence)				\
       T(AsyncSequence)				\
       T(ReactiveSequence)			\
       T(SequenceWithMemory)			\
       T(Fallback)				\
       T(ReactiveFallback)			\
       T(AsyncFallback)				\
       T(Parallel)				\
       T(ParallelAll)				\
       T(RetryUntilSuccessful)			\
       T(IfThenElse)				\
       T(WhileDoElse)				\
       T(Inverter)				\
       T(PipelineSequence)			\
       T(KeepRunningUntilFailure)		\
       T(RoundRobin)				\
       T(Recovery)				\
       T(RateController)			\
       T(ForceSuccess)				\
       T(ForceFailure)				\
       T(Repeat)				\
       T(AlwaysSuccess)				\
       T(AlwaysFailure)				\
       T(SubTree)

#define T(x) x,
typedef enum { BTTYPES BTYPES_MAX }  BTNode_Type;
#undef T

#define BTRET							\
  T(SFR)							\
       T(SF)							\
       T(SR)							\
       T(FR)							\
       T(S)							\
       T(F)							\
       T(R)

#define T(x) x,
typedef enum { BTRET BTRET_MAX }  BTReturn_Type;
#undef T

typedef enum { NODE,		/* One Fiacre/hippo tick for every nodes (the default) */
	       ACTION,		/* Only for action */
	       BT}		/* One per top level BT */
  HTick_Type;

extern HTick_Type htick;

typedef Slist *BTNodeList;

typedef struct BTNode_structure BTNode;

struct BTNode_structure {
  BTNode_Type type;
  Symbol name;
  PString file_name;
  Symbol id;
  //  BTNode *child;
  BTNodeList children;
  Slist *args;
  BTReturn_Type PossibleRet;
  PBoolean pass_arg;
  PBoolean wait;		/* used for // */
  PBoolean halt;		/* used for // */
  int tcldot;
  Symbol sv;
  Expression *expr;
  union {
    int num_attempts;
    int repeat;
    int needed_success;		/* for Parallel nodes */
  } u;
};

typedef struct {
  Symbol id;
  Sprinter *arg_string;
}  BTId_Action_Arg;


extern Slist *bt_list;	
extern Slist *sv_list;
extern Slist *btnode_list;
extern Slist *btidactarg_list;

extern Slist *btn_stack;	/* used during the parking */

extern int bt_index; 		/* First available node tcldot index */

//extern BTree *current_bt;

struct event {
  Symbol name;
  Expression *guard;
  ExprList effects;
};
  
struct sv {
  Symbol name;
  SymList states;
  Symbol init;
  PBoolean all_transitions;
  Slist *transitions;
  int initi;
  int max;
  int min;
};

struct res {
  Symbol name;
  int init;
  int max;
};

struct transition {
  State from;
  State to;
};


typedef struct key_val {
  Symbol key;
  Term *val;
} KeyVal;


struct btset {
  Symbol name;
  Slist *bt_list;
  Slist *res_list;
  Slist *event_list;
  Slist *skill_list;
};

extern Event *current_event;
extern Sv *current_sv;
extern Res *current_res;
// extern BTNode *current_btnode;

extern long ns_per_tick;

typedef struct time_interval Time_Interval;

struct time_interval {
  int min;
  int max;
  PBoolean max_inf;
};

  
struct skill {
  Symbol name;
  Slist *precondition;
  Slist *invariant;
  PBoolean monitor;
  ExprList start;
  ExprList interrupt;
  Time_Interval *time_interval;
  Slist *failure;
  Slist *success;
  Op_Structure *op; 		/* This really contains the OP body and action... */
  Expression *progress;
  Slist *dbp;
  PBoolean pass_arg;
  Sprinter *arg_string;
  int index_level[10];		/* more than 10 level deep? unlikely... */

};

struct skillset {
  Symbol name;
  Slist *sv_list;
  Slist *res_list;
  Slist *event_list;
  Slist *skill_list;
};

#endif /* INCLUDE_bt_structure */
