/*                               -*- Mode: C -*- 
 * bt-structure_f.h -- Declaration of external functions of bt-structure.c
 * 
 * Copyright (c) 1991-2024 Francois Felix Ingrand.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

KeyVal *make_key_val(Symbol key, Term *val);

Symbol btnode_name(BTNode *btnode);
PString btnode_file_name(BTNode *btnode);


void init_make_btnode(Symbol name, PBoolean graphic);
void print_btnode(BTNode *btnode);

BTNode *make_btnode();
BTNode *make_btnode_type(BTNode_Type type);
BTNode* current_btnode();

void set_needed_success_if_not_set(BTNode *n);

//BTree *make_bt();

BTNode *dup_btnode(BTNode *btnode);
void free_btnode(BTNode *btnode);
void free_node_from_free_btnode(Node *node);
void free_edge_from_free_btnode(Edge *edge);

PBoolean sort_btnode(BTNode *btnode1, BTNode *btnode2);

Skill *make_skill(Symbol name);
SkillSet *make_skillset(Symbol name);
Sv *make_sv(Symbol name);
Res *make_res(Symbol name);
Event *make_event(Symbol name);
Transition *make_transition(State from, State to);
Precondition *make_precondition(Symbol name, Expression *expr);
Time_Interval *make_time_interval(int min, int max, PBoolean max_inf);
Invariant *make_invariant(Symbol name);
Failure *make_failure(Symbol name);
Success *make_success(Symbol name);
void add_sv_to_sv_list(void);
void add_res_to_res_list(void);
void add_skill_to_skill_list(void);
void add_event_to_event_list(void);

PBoolean write_fcr(PString file_name);

void  finish_loading_btnode(BTNode *btnode);

