/*                               -*- Mode: C -*- 
 * opaque.h -- Opaque declarations
 * 
 * Copyright (c) 1991-2024 Francois Felix Ingrand.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef INCLUDE_opaque
#define INCLUDE_opaque


#include "slistPack.h"
#include "slistPack_f.h"
#include "shashPack.h"

/* 
 * An opaque pointer is, most of the time, a pointer to a structure not 
 * yet defined. AS slong as you do not dereference ce structure, 
 * that's OK, the compiler is able to handle this opaque pointer.
 * 
 */

typedef char PBoolean;

typedef unsigned char Uchar;

typedef const char *Symbol;	/* Symbol are "fixed" pointers. */

typedef struct pred_rec Pred_Rec;
typedef struct func_rec Func_Rec;

typedef char *PString;

typedef Slist *TermList;	       
typedef Slist *Intention_List;	       
typedef struct term Term;
typedef struct term *TermPtr;
typedef TermPtr (*PFPT)(TermList); /* PFPT is pointer to a function taking a TermList and returning a
				      pointer to a Term... I love C decl.. */

typedef Slist *(*PFPL)(Slist *); /* PFPL is pointer to a function taking a Slist *and returning a
				      pointer to Slist *.*/

/* Pointer on function returning void... */
typedef void (*PFV)();		

typedef void entrypoint(void);
typedef void entrypoint_name(char *);

typedef struct expression Expression;

typedef Slist *VarList;

typedef struct envar Envar;

typedef struct l_list *L_List;

typedef struct int_array Int_array;
typedef struct float_array Float_array;

typedef struct eval_pred Eval_Pred;

typedef struct eval_funct Eval_Funct;

/*  pointer to a function returning a PBoolean */
typedef PBoolean (*PFB)(); 
/*  pointer to a function returning a PBoolean, taking  TermList*/
typedef PBoolean (*PFBPTL)(TermList);

typedef struct bt2f Bt2f;
typedef struct thread_intention_block Thread_Intention_Block;
typedef struct intention Intention;
typedef struct op_instance Op_Instance;
typedef struct intention_graph Intention_Graph;

typedef struct fact Fact;
typedef struct goal Goal;
typedef struct u_memory U_Memory;

typedef struct frame *FramePtr;

typedef Slist *ExprList;		/* Les elements sont des Expression */
typedef Slist *ExprFrameList;	/* les elements sont des pairs, expression, frame */
typedef Slist *ExprListFrameList; /* les elements sont des pairs, list d'expression, frame */
typedef Slist *FrameList;
typedef Slist *TermTypeList;

typedef struct draw_data Draw_Data;
/* typedef struct gnode Gnode; */
/* typedef struct gedge Gedge; */
/* typedef struct gedge_text Gedge_text; */
/* typedef struct gtext Gtext; */
/* typedef struct int_draw_data Int_Draw_Data; */
/* typedef struct og OG; */
typedef Slist *List_OG;
typedef struct iog IOG;
typedef Slist *List_IOG;

typedef struct op_structure Op_Structure;

typedef struct control_point Control_Point;
typedef struct control_point Node;

typedef struct logic Logic;
typedef struct logic Edge;

typedef Slist *Subst_List;
typedef Slist *Op_List;
typedef Slist *Op_Instance_List;


typedef Slist *Condition_List;

typedef Slist *Op_Expr_List;

typedef Slist *CpList;		/*Control point list */
typedef Slist *NodeList;

typedef Slist *EdgeList;
typedef Slist *LogicList;

typedef Slist *Fact_List;
typedef Slist *Goal_List;

typedef Slist *BE_Clean_List;

typedef Slist *ListLines;

typedef struct oprs_client Oprs_Client;

typedef Slist *Thread_Intention_Block_List;

typedef Slist *Intention_Paire_List;	/* List de paire d'intentions */

/* Pointer on function returning int... */
typedef int (*PFI)();		

/* Pointer on function returning pointer on void... */
typedef void * (*PFPV)();		

/* Pointer on function returning OPRS_NODE... */
/* typedef OPRS_NODE (*PFN)();		 */

/* Pointer on function returning Intention_List. */
typedef Intention_List (*PFIL)();

/* Pointer on function returning ListLines. */
typedef ListLines (*PFLL)();

typedef struct database Database;

typedef struct relevant_op Relevant_Op;

typedef struct body Body;

typedef struct key_elt Key_Elt;

typedef struct key Key;

typedef Slist *SymList;

typedef struct type Type;

typedef Slist *TypeList;

typedef struct skill Skill;

typedef struct event Event;

typedef struct sv Sv;

typedef struct res Res;

typedef struct skillset SkillSet;

typedef Symbol State;

typedef struct transition Transition;

typedef struct precondition Precondition;

typedef struct invariant Invariant;

typedef struct success Success;

typedef struct failure Failure;

typedef struct time_interval Time_Interval;

#endif /* INCLUDE_opaque */
