#!/usr/bin/wish
package require Tk
package require Tcldot

# from https://wiki.tcl-lang.org/page/Scrolled+frame%2C+canvas%2C+auto+scrollbars
proc scrolled w {
    set pa [set pa0 [winfo parent $w]]
    if {$pa eq "."} {set pa ""}
    grid $w [scrollbar $pa.y -command "$w yview"] -sticky news
    $w configure -xscrollcommand "$pa.x set" -yscrollcommand "$pa.y set"
    grid $pa.y -sticky ns
    grid [scrollbar $pa.x -ori hori -command "$w xview"] -sticky ew
    grid columnconfigure $pa0 0 -weight 1
    grid rowconfigure    $pa0 0 -weight 1
}


# from https://wiki.tcl-lang.org/page/Canvas+zooming+using+mousewheel

proc zoominit {c {zfact {1.1}}} {
        # save zoom state in a global variable with the same name as the canvas handle
        upvar #0 $c data
        set data(zdepth) 1.0
        set data(idle) {}
        # add mousewheel bindings to canvas
        bind $c <Button-4> "zoom $c $zfact"
        bind $c <Button-5> "zoom $c [expr {1.0/$zfact}]"
        bind $c <MouseWheel> "if {%D > 0} {zoom $c $zfact} else {zoom $c [expr {1.0/$zfact}]}"
}

proc zoom {c fact} {
        upvar #0 $c data
        # zoom at the current mouse position
        set x [$c canvasx [expr {[winfo pointerx $c] - [winfo rootx $c]}]]
        set y [$c canvasy [expr {[winfo pointery $c] - [winfo rooty $c]}]]
        $c scale all $x $y $fact $fact
        # save new zoom depth
        set data(zdepth) [expr {$data(zdepth) * $fact}]
        # update fonts only after main zoom activity has ceased
        after cancel $data(idle)
        set data(idle) [after idle "zoomtext $c"]
}
                                                                                                                                                         
proc zoomtext {c} {
        upvar #0 $c data
        # adjust fonts
        foreach {i} [$c find all] {
                if { ! [string equal [$c type $i] text]} {continue}
                set fontsize 0
                # get original fontsize and text from tags
                #   if they were previously recorded
                foreach {tag} [$c gettags $i] {
                        scan $tag {_f%d} fontsize
                        scan $tag "_t%\[^\0\]" text
                }
                # if not, then record current fontsize and text
                #   and use them
                set font [$c itemcget $i -font]
                if {!$fontsize} {
                        set text [$c itemcget $i -text]
                        if {[llength $font] < 2} {
                                #new font API
                                set fontsize [font actual $font -size]
                        } {
                                #old font API
                                set fontsize [lindex $font 1]
                        }
                        $c addtag _f$fontsize withtag $i
                        $c addtag _t$text withtag $i
                }
                # scale font
                set newsize [expr {int($fontsize * $data(zdepth))}]
                if {abs($newsize) >= 4} {
                        if {[llength $font] < 2} {
                                #new font api
                                font configure $font -size $newsize
                        } {
                                #old font api
                                set font [lreplace $font 1 1 $newsize] ; # Save modified font! [ljl]
                        }
                        $c itemconfigure $i -font $font -text $text
                } {
                        # suppress text if too small
                        $c itemconfigure $i -text {}
                }
        }
        # update canvas scrollregion
        set bbox [$c bbox all]
        if {[llength $bbox]} {
                $c configure -scrollregion $bbox
        } {
                $c configure -scrollregion [list -4 -4 \
                        [expr {[winfo width $c]-4}] \
                        [expr {[winfo height $c]-4}]]
        }
}

set c [canvas .c -background beige -width 1500 -height 800]
pack $c
zoominit .c
#set f [open "ros2-nav2.dot" r]
puts $dot_filename
set f [open $dot_filename r]
set g [dotread $f]
if {[info exists debug]} {
    puts [$g render]         ;# see what render produces
}
eval [$g render]
.c configure -scrollregion [.c bbox all]
scrolled .c
